#this file describes the wrapper for reading files

#this might not need it's own file lmao (this is an old comment)

import csv

#this function takes a string and splits it into each instruction and its arguemnts
    #this is done by making use of csv reader which does this automatically
    #it could be debated if this should be moved into parse, but in my opinion this is more of a
    #lexer than a parser but having an entire file just for this function is a bit unecersarry
def read_str(string):
    tmp = [n for n in csv.reader(string.split("\n"), delimiter=" ")]
    return tmp



#this file describes the api for the bytecodecompile package, which takes a string of bytecode in and returns a code object out
    #it also contains the main function which is a relic from when this project was different

import marshal
import sys
import os

#my code imports
import bytecodemacro.bytecodecompiler.Read as Read
import bytecodemacro.bytecodecompiler.Parse as Parse
import bytecodemacro.bytecodecompiler.PreProcess as PreProcess
import bytecodemacro.bytecodecompiler.CodeContext as CodeContext
import bytecodemacro.bytecodecompiler.Compile as Compile

#this function takes a string of the modified bytecode and converts it to a python code object and
def string_to_code(string, one_obj=False):
    read = Read.read_str(string)
    all_objs = [[Parse.get_args(n)] for n in Parse.split(read)] #objects are mapped to array so that we can use the pass by reference nature of the array to handle objects referencing other objects
    [PreProcess.PreProcess(n[0], all_objs) for n in all_objs] 
    for n in all_objs:
        n[0] = n[0].MakePyObj()
    if one_obj:
        if len(n) != 1: raise Exception("one_obj true but more than one object passed")
        return n[0]
    else:
        return next(n[0] for n in all_objs if n[0].co_name.upper() == "MAIN") #gets the object named main and returns it



#this file takes code context in with a string bytecode and converts it to a list of tuples that are
    #properly formatted with byte arguments as opposed to the previous string arguments

from functools import reduce, partial
from bytecodemacro.bytecodecompiler.CodeContext import CodeContext

#this function takes in a code object and a list of all the code objects and returns the same code
#object with a processed bytecode and attributes that match (like the consts array holds constants).
    #the reason it is a massive closure is because due to the nature of the bytecode, writing this
    #with state makes more sense and as this function can be called multiple times it's better to
    #keep the state in the function so that it is remade and not kept the same between objects
    #although this function is quite ugly as far as coding goes, especially with all that state.
    #hopefully there is a better way to do this but I can't think of one without seriously downsides
def PreProcess(code_context_in, all_objects):

    #the naming schema is a bit weird with these lists but it is to keep it consistent with the
    #documentations (in the dis library) names
    constants = [] #list of literals used in the program (i would use a set for faster seraching but the instructions rely on the index of the literal so i have to)
    names = [] #list of names used in the program (same logic about sets as above)
    varnames = code_context_in.varnames #(the arguments already exist as varnames)
    global_list = [] #although in documentation this is globals but I can't name it that due to globals being reserved
    cellvars = [] #not sure what this is (this is the freevars storage but i'm not sure how to do this)

    compare_list = ('<', '<=', '==', '!=', '>', '>=') 
    label_positions = {}

    #used to compose the pre-processing functions together
        #compose and bind are both used in both pre_processing here and uncompile and should be
        #moved somewhere else to increase code reuse
    def compose_list(f_list):
        def ret(arg):
            return reduce(lambda a,n: n(a), f_list, arg)
        return ret

    #equivalent to the list monads bind operator. any attempt to explain this will end up just
    #re-writing the code in comment form, please try to understand this be reading it yourself
    def bind(f, li): #f is a function that takes an element from li and reutrns a list
        return list(reduce(lambda a1,a2: a1+a2, map(f, li))) #the + in here is list concatenation

    #this function takes a line and returns a boolean if it should be removed from the bytecode
    def should_remove(line): #called in the filter function
        if len(line) == 0: return False #this is already done in parse but i've kept it here just in case i delete it there
        elif line[0][0] == "#": return False #manual comments in the code, note: this could be deleted
        else: return True

    #this function finds all of the labels, changes them to nop and adds there positions to label_positions
    def find_labels(lines): #not a massive fan of this function but it works
        ret = []
        for i, v in enumerate(lines):
            if len(v) == 2 and v[0] == "LABEL":
                label_positions[v[1]] = i*2 #*2 because it is the position in the byte string and
                                              #it is 2 bytes per instruction
                v = ["NOP"]
            ret+=[v]
        return ret

    #removes any empty arguments in a tuple
    def clean_args(val):
        if len(val) == 1: return val
        if val[1] == '':  return [val[0]]
        else:             return val[:2]

    #for instrucitons that don't have arguments (like pop_top or return_value)
        #note: this is not likely to occur in the context of a macro
    def add_arg(val):
        if len(val) == 1: 
            return val + [0]
        return val

    #when an instruction is called that its argument is an index to a list attribute in the original
    #bytecode, this function is used to add that argument to the attribute in the new bytecode when
    #we are generating it, this also prevents the same argument from getting added to the bytecode
    #attributes more than once
    #e.g: load_const "hello world" will add hello world to the constants array and return load_const 1
    def memoize_loading(lis, arg): #note: memoize loading is called sometimes with things that should not be added to (what does this mean?)
        try: 
            arg = int(arg)
        except ValueError:
            try:
                arg = float(arg)
            except ValueError: pass
        finally:
            if arg in lis: arg = lis.index(arg) #double check later, also very slow
            else: lis.append(arg); arg = len(lis)-1 #maybe remove semi colon but I like how this looks
            return (lis, arg)

    #commands to lists is used in handle_literals
        #note: i am not sure how load_closurer works and it has something weird about freevars in it
    commands_to_lists = {"LOAD_CONST": constants, "LOAD_NAME": names, "LOAD_GLOBAL":
            global_list, "LOAD_FAST": varnames, "STORE_FAST": varnames, "DELETE_FAST":
            varnames, "LOAD_ATTR": names,
            "LOAD_CLOSURE": cellvars, "LOAD_DEREF": cellvars, "LOAD_CLASSDEREF": cellvars,
            "DELETE_DEREF": cellvars, "LOAD_METHOD": names, "STORE_NAME": names,
            "STORE_ATTR": names, "STORE_GLOBAL": global_list, "STORE_DEREF": cellvars,
            "DELETE_NAME": names, "DELETE_ATTR": names, "DELETE_GLOBAL": global_list,
            "IMPORT_NAME": names, "IMPORT_FROM": names, "COMPARE_OP": compare_list}
    #this function will take an instructino and if it contains an instruction that would normally
    #index into an attribute of an array in the original bytecode, it calls memoize_loading to handle it
    def handle_literals(val):
        inst, arg = val
        if inst in commands_to_lists:
            lis = commands_to_lists[inst]
            lis, arg = memoize_loading(lis, arg)
        return (inst, arg)

    #this function takes a line and if it is a jump absolute literal (with an absolute position in the
    #byetcode), and converts the second argument into the correct position in the code rather than the label
        #references a dictionary not a list so it has to be done like this (what does this mean?)
    def handle_jump_absolute_literals(line):
        inst, arg = line
        if inst in {"POP_JUMP_IF_TRUE", "POP_JUMP_IF_FALSE", "JUMP_IF_NOT_EXC_MATCH", "JUMP_IF_TRUE_OR_POP", "JUMP_IF_FALSE_OR_POP", "JUMP_ABSOLUTE"}:
            arg = label_positions[arg]
        return (inst, arg)

    #obj_dict is useful in handle_load_object
    obj_dict = {o[0].name: o for o in all_objects}
    #this function handle the custom instruciton load_object
    #it takes a line in and if that line is load_object it will load the constant object in
        #this function is not too useful in a macro, so it might have to be deleted in a later version
    def handle_load_object(line): 
        nonlocal constants
        inst, arg = line
        if inst == "LOAD_OBJECT":
            inst = "LOAD_CONST" #how this works is it first loads the object as a constant, but this
                                #object is in a one long array, as to emulate pointer behaviour in python to allow for recursive objects
                                #it also makes compilation of the objects a lot easier
            constants, arg = memoize_loading(constants, obj_dict[arg])
            inst2 = "LOAD_CONST" #it then loads 0 onto the stack
            constants, arg2 = memoize_loading(constants, 0) #i should be calling handle_literals instead of this, but tbh at this point i just don't care
            return [(inst, arg), (inst2, arg2), ("BINARY_SUBSCR",0)] #it then subscripts into the one length array to extract the python code object onto the stack
        return [(inst, arg)]

    #this function handles jumps statements that their arguments are relative jumps
    def handle_jump_relative_literals(lines):
        ret = []
        for i, v in enumerate(lines):
            inst, arg = v
            if inst in {"SETUP_WITH", "JUMP_FORWARD", "FOR_ITER", "SETUP_FNIALLY"}:
                arg = label_positions[arg] - i*2 #i*2 as there are two bytes per instruction
            ret += [(inst, arg)]
        return ret

    #used in debugging 
    def trace(v):
        print(v)
        return v

    #sometimes instructions need to be added after the labels positions have been defined in the
    #label position object so this function updates those labels to be in the correct spot
    def increment_labels_below(index, n):
        nonlocal label_positions
        label_positions = {key: val if val < index else val + n*2 for key, val in label_positions.items()} #n * 2 as instructions are 2 bytes long

    #this function checks to see if any argument in the tuple is above 255 and if it is it adds
    #extended arg instructions to ensure that the arguments are always byte sized
        #terrible name for a function
    def extended_args_instructions(n): #i like this functions implementation :)
        if n < 256: return [("EXTENDED_ARG", n)]
        if n > 255: 
            tmp = ("EXTENDED_ARG", n % 256)
            return [tmp] + extended_args_instructions(n >> 8)

    #this function takes in the byte tuples and if any argument is above byte sized, it adds
    #extended args instructions to the bytecode.
        #this function does none of the maths but it calls the extended_args_instruction function
        #which handles all of the hard work
    def handle_extended_args(lines):
        ret = []
        for i, v in enumerate(lines):
            inst, arg = v 
            if arg > 255: 
                e = extended_args_instructions(arg >> 8)
                increment_labels_below(i, len(e))
                arg = arg % 256
                ret += e
            ret.append((inst, arg))
        return ret

    #this creates a function pre_process that pre_processes a bytecode tuple
        #maybe not the cleanest to do it like this but it's more fun like this (i made this before I
            #knew other people were going to see this), it is probably worth changing as it's not the most readable
        #the order of the functions here is not completely arbitrary and any changes to this order
            #must be properly thought out
    pre_process = compose_list([
        partial(filter, should_remove),
        find_labels, #this can't be in the map cause it needs to have index information
        partial(map, compose_list([
            clean_args,
            add_arg, 
            lambda n: (n[0].upper(), n[1]), #makes the instructions uppercase (has to be a lambda cause .upper is a method (ruins my pointfree >:(
            handle_literals,
            handle_jump_absolute_literals
        ])), #composes all the preproccessing that only needs a single line and not the context of other lines
        partial(bind, handle_load_object),
        handle_jump_relative_literals,
        partial(map, lambda n: (n[0], int(n[1]))),
        handle_extended_args,
        list #list might not be necersarry, but it's better safe than sorry in future updates
    ])

    #this section of the code modifies all of the attributes of the code_context to their properly pre_processed version
    code_context_in.bytes = pre_process(code_context_in.bytes)
    code_context_in.constants = constants
    code_context_in.names = names
    code_context_in.global_list = global_list
    code_context_in.cellvars = cellvars
    code_context_in.varnames = varnames



#this file describes the api for the uncompile library
    #it also contains some unecersarry code (in the main function) from when this was a different
    #project which should be removed at a later date

import bytecodemacro.uncompile.dis_wrapper as dis_wrapper
import bytecodemacro.uncompile.parse_dis as parse_dis
import bytecodemacro.uncompile.pre_process as pre_process
import bytecodemacro.uncompile.obj_handler as obj_handler
from functools import reduce

#this function takes in a python code object and returns the bytecode tuple representation
    #this is done in multiple steps for dbeugging and clarity
def main_tups(obj):
    objs = obj_handler.get_objects(obj)
    dis_strs = [dis_wrapper.get_str(n) for n in objs]
    parsed = [parse_dis.parse(n) for n in dis_strs]
    pre_processed = [pre_process.pre_process(n) for n in parsed]
    out_tuples = [obj_handler.add_definitions(objs[n], pre_processed[n]) for n in range(len(pre_processed))]
    out_tuples = list(reduce(lambda l1,l2: l1+l2, out_tuples))
    return out_tuples #not actually objects but lists of tuples with object information

#this function takes a list of tuples and returns them as a single concatenated string
    #although this probably should be in a separate file it was unecersarry and useful to expose globally
    #it could even be moved to bytecodecompile
def tups_to_str(tuples):
    out = ""
    for v in tuples:
        for i in v:
            out += i + " "
        out = out[:-1] #remove the extra space
        out += "\n"
    out = out[:-1] #remove extra new line
    return out



#this file describes the api that the user can access through the bytecode macro library

import bytecodemacro.bytecodecompiler.Runner as runner
import bytecodemacro.uncompile.main as main
from types import FunctionType

#this is the main entry point for the api. It's a curried function that takes two functions and uses
    #the first to modify the seconds bytecode and returns the newly created function from that.
def macro(mac_mod):
    def ret(f):
        tups = main.main_tups(f.__code__)
        new_func = mac_mod(tups)
        string = main.tups_to_str(new_func)
        code = runner.string_to_code(string, one_obj=True)
        return FunctionType(code, globals())
    return ret

#this takes a string and compiles it to bytecode tuples
    #relies on no return statements being sent
    #only works with single statements
def byte_compile(string): #need to pass the state inside the funcdtion with like variables and that
    obj = compile(string, "string", "single") #here
    tups = main.main_tups(obj)
    tups = tups[1:-3] #removes define and load_const None return_value end
    return tups

#this function is currently unused and untested
    #it is the same as byte_compile except it returns the definie and return_value statement
    #for when you have a new definition created in the string
def byte_compile_obj(string):
    obj = compile(string, "string", "exec")
    tups = main.main_tups(obj)
    return tups

#this function takes some tuples, compiles them and then runs thme
    #only works with one object
def execute_tups(tups):
    string = main.tups_to_str(tups)
    code = runner.stirng_to_code(string, one_obj=True)
    exec(code)

#this function takes two lists of tuples and returns a new variable name that is unused in both of the lists
    #quite slow and could be optimized later but this is good enough for now
def unused_var(old_tups, new_tups, test="v1", am=1):
    for n in old_tups + new_tups:
        if len(n) == 2:
            inst, arg = n
            if arg == test: return unused_var(old_tups, new_tups, "v" + str(am+1), am+1)
    return test

#maybe add unused jump name
    #same as unused_var except for jump names



#this file takes a string returned by the dis wrapper and parses it into tuples
    #these tuples are not pre processed and as such are not always only 2 long and also does not
    #contain information that the objects hold
#the entry point for this file is def parse(obj_str):

#given a string return the position of the next whitespace
def find_next_whitespace(string):
    pos = 0
    for n in string:
        if n == " ": return pos
        pos+=1
    raise Exception("find next whitespace called without any whitespace")

#given a string return the position of the next non_whitespace
def find_non_whitespace(string):
    pos = 0
    for n in string:
        if n != " ": return pos
        pos += 1
    raise Exception("find non whitespace called with only whitespace")

#this function takes in a line from the dis string and splits it into it's arguments
    #including the third argument which is the dis library printing what the third arguments literal is
    #i.e: 0         load_const 0 ("hello world") could be a line that gets passed
def split_line(line):
    useable = line[16:] #removes the line numbers and pointer things
    if len(useable) < 25: return (useable,) #useable must be only one instruction
    w = find_next_whitespace(useable) #could replace with useable.index(" ")
    inst = useable[:w]
    arg = useable[w:]
    nw = find_non_whitespace(arg)
    arg = arg[nw:]
    if "(" not in arg:
        return (inst, arg)
    n = arg.index("(")
    n2 = arg.index(")") #these are causing errors with strings that have brackets in them
    arg_val = arg[n+1:n2]
    arg_num = arg[:n-1]
    return (inst, arg_num, arg_val) #might not need to return arg_num but whatever

#takes in the string returned by the dis and returns the un processed tuples from the string
def parse(obj_str):
    lines = obj_str.split("\n")
    args = [split_line(n) for n in lines]
    return args



#this file describes the wrapper used to access the dis library
    #this file is unecersarry and should be removed at a later date

import dis

#gets the string output of dis from the py_object supplied
def get_str(py_object):
    tmp = dis.Bytecode(py_object)
    return tmp.dis()



#this file describes the functions that handle code objects when they are used in the uncompile library

import types

#this function gets all the code objects from the constants array of another code object
    #this is useful for when a file is compiled and it contains many code objects, it may not be necersarry anymore
    #update to add inner function and class support (idk about classes i just haven't tested it)
def get_objects(obj): 
    ret = [n for n in obj.co_consts if isinstance(n, types.CodeType)]
    return [obj] + ret

#this function gets all of the argument names from an object
def arg_names(obj):
    return [obj.co_varnames[n] for n in range(obj.co_argcount)]

#this function takes a code object and it's bytecode tuple representation and outputs the header
#define and add_arg statements and the end statement to make it properly structured
def add_definitions(obj, tuples):
    out = [("DEFINE", "main" if obj.co_name == "<module>" else obj.co_name)]
    out += [("ADD_ARG", n) for n in arg_names(obj)]
    out += tuples
    out += [("END", "0")]
    return out



#this file takes a list of tuples (with mnenoic instructions) and converts it to a bytes.
    #it also contains code to write out a file with a header that is a relic of when this was a
    #different project

import dis

#compiles out a byte string into a bytes
def Compile(BytesStr):
    ret = []
    for n in BytesStr:
        ret += [dis.opmap[n[0]], n[1]]
    return bytes(ret)



import pathlib
from setuptools import setup, find_packages

HERE = pathlib.Path(__file__).parent

README = (HERE / "README.md").read_text()

setup(
        name="bytecodemacro",
        version="2.0.4", #whoops accidentally did too many versions
        description="a macro library that expoes the bytecode of python",
        long_description=README,
        long_description_content_type="text/markdown",
        url="https://github.com/Glubs9/bytecodemacro",
        author="glubs9",
        author_email="jonte.fry@gmail.com", #please no spam
        license="GPLv3",
        #classifiers=[ #apparnetly this is not a valid classifier idk how this works
        #    "Programming Language :: Python :: 3.9",
        #],
        packages=find_packages() #can't be bothered to do properly
        #i'm not too sure how to specify entry_points (or if i need to)
)



#this file takes the unrefined tuples from parse and returns the tuples properly pre_processed of 2 length tuples
    #the entry point for this file is the function pre_proccess (at the bottom of the file)

from functools import reduce

#this function is an extremely useful utility function that is equivalent to the bind operation on a list monad
    #here is the type signature for this function (a -> [b]) -> [a] -> [b] (in haskell syntax)
    #any attempt to explain what this does will just explain the code directly so I will just hope that you can read it
def bind(f): #f takes an item from the list and returns an array
    def ret(li):
        return reduce(lambda n1,n2: n1+n2, map(f, li)) #reduce concatenates the arrays together
    return ret

#compose takes an inp and a series of functions, it then calls the functions as such: f1(f2(...fn(inp)))
def compose(inp, *fs):
    for n in fs:
        inp = n(inp)
    return inp

#this function tests if an instruciton loads a code object
test_str = "<code object " #not the most rigorous check but I think it will work for now
def is_load_object(v):
    return v[0] == "LOAD_CONST" and v[2][:len(test_str)] == test_str

jump_statements = {"POP_JUMP_IF_TRUE", "POP_JUMP_IF_FALSE", "JUMP_IF_NOT_EXC_MATCH", "JUMP_IF_TRUE_OR_POP", "JUMP_IF_FALSE_OR_POP", "JUMP_ABSOLUTE", "SETUP_WITH", "JUMP_FORWARD", "FOR_ITER", "SETUP_FNIALLY"}

from bytecodemacro.uncompile.parse_dis import find_next_whitespace #this is placed here because it is only used by handle_third arg
#this function takes in an un_processed tuple with potentially a third argument, converts that
#argument to the modified bytecode syntax and returns it
    #todo: handle extended_arg instruction (i think we just delete it tbh, idk i've never seen it in compiled python)
def handle_third_arg(v):
    if len(v) == 1: return [(v[0], '0')]
    elif len(v) == 2: return [v]
    inst = v[0]
    if inst in {"SETUP_WITH", "JUMP_FORWARD", "FOR_ITER", "SETUP_FNIALLY"}:
        _, arg = v[2].split() #v[2] == "to x" where x is the absolute jump position
        return [(inst, arg)]
    elif inst in jump_statements:
        raise Exception("these should only be 2 long")
    elif is_load_object(v):
        #if this is load_const with <code object test at>
        #replace with load_object
        arg = v[2][13:] #gets the name of the object
        arg = arg[:find_next_whitespace(arg)]
        inst = "LOAD_OBJECT" #not techincally a python bytecode thing but it is in the modified bytecode
        return [(inst, arg)]
    elif inst in {"LOAD_CONST", "LOAD_NAME", "LOAD_GLOBAL", "LOAD_FAST", "STORE_FAST", "DELETE_FAST", "LOAD_ATTR", "LOAD_CLOSURE", "LOAD_DEREF", "LOAD_CLASSDEREF", "DELETE_DEREF", "LOAD_METHOD", "STORE_NAME", "STORE_ATTR", "STORE_GLOBAL", "STORE_DEREF", "DELETE_NAME", "DELETE_ATTR", "DELETE_GLOBAL", "IMPORT_NAME", "IMPORT_FROM", "COMPARE_OP"}:
        return [(inst, v[2])]#need to double check each one of the load_statements to make sure am parsing properly but i'm pretty sure this works

#takes in the pre_processed lines and adds labels to where the jump statements jump (as jump statement jump by line number not label in bytecode)
def handle_jumps(lines):
    jump_lines = set()
    for v in lines: #could be done with a set comprehension but this is clearer
        inst, arg = v
        if inst in jump_statements: 
            jump_lines.add(int(arg)) #found a place where the jump jumps to and adds it to the set
    ret = []
    for i in range(len(lines)):
        if i*2 in jump_lines: #i*2 because the jump statements are made by counting bytes not liens
            ret.append(("LABEL", str(i*2))) #the label is named like this so I don't have to change the jump statements
        ret.append(lines[i])
    return ret

#this function returns an empty list if the line is empty, but is identity for non_empty lines
    #this function is called in a bind, so in the bind it removes the empty lines
def remove_empty(line):
    if line[0] == '' and len(line) == 1: return []
    return [line]

#strings are given with ' ' but my cpyasm only works with " " so this function replaces any instances of the ' in the line passed
    #this can cause some issues when passing strings from macro_lib so it needs to be fixed later
def handle_strings(v):
    inst, arg = v
    if arg[0] == "'":
        arg = '"' + arg[1:-1] + '"'
    return (inst, arg)

#if the line passed is a load_global replace with a load_name
    #i'm not 100% on what this does but i'm pretty sure it fixed some bug somewhere
    #take a look at this in the future
def remove_globals(v):
    inst, arg = v
    if inst == "LOAD_GLOBAL":
        inst = "LOAD_NAME"
    return (inst, arg)

#for debugging, it is not currently used but it's really annoying to write again all over the shop so we live it like this for now
def trace(msg):
    def ret(n):
        print(msg + ": " + str(n))
        return n
    return ret

#this is the main entry point for the file
#it takes in a list of instructions, as returned by parse and returns a pre+processed and properly
#structured verison of those bytecode. 
    #note: this function does not handle the object structure (define, add_arg, end) which is handled in obj_handler.py
def pre_process(instructions):
    return compose(instructions, #if we changed compose to be curried this function could be pointfree but that would be less readable and less pythonic
        bind(remove_empty),
        bind(handle_third_arg),
        lambda n: map(handle_strings, n), #if only map was curried :(
        lambda n: map(remove_globals, n), #this is just cause i'm too lazy to handle globals ec dee (this is an old comment and might not be true on the current build)
        list, #this is because map does not reutrn a list but an iterable but handle_jumps requires a list
        handle_jumps
    )



#this file splits the input string into a list of objects,
    #it also has code for getting the arguments of an object which could be moved to PreProcess but
    #I am unsure about it

from bytecodemacro.bytecodecompiler.CodeContext import CodeContext

def split(lines):
    """ 
    split splits the files into a series of lines for each object
    (this takes already lexed information from the csv reader)
    """
    curr_obj = None #gets defined as a CodeContext
    all_objs = []
    for n in lines:
        if len(n) == 0 or n[0][0] == "#": continue #this is preprocessing which should be done in preprocess.py but
                                 #i'm doing it here because the parser requires thsi to functoin
        elif n[0].upper() == "DEFINE":
            #remove empty lines here to properly work
            args = [i for i in n[2:] if i != ""]
            curr_obj = CodeContext(name=n[1], arg_num=len(args), varnames=args)
        elif n[0].upper() == "END":
            all_objs += [curr_obj]
            curr_obj = None
        else:
            curr_obj.bytes += [n]
    return all_objs

#this takes an object, adds the arguments to the object, then returns it
    #this should be in PreProcess but i put it in here because ?? (i mean it is small and elegant and i don't want to look at PreProccess rn tbh)
    #note: although this function does change an object, which is passed by reference, please make use of this function as if it passed objects by value
def get_args(obj):
    tups = obj.bytes
    n = 0
    args = []
    while tups[n][0] == "ADD_ARG":
        args += tups[n][1]
        n+=1
    obj.varnames = args
    obj.arg_num = len(args)
    obj.bytes = obj.bytes[n:] #removes add_arg from the bytes
    return obj
